// Copyright 1997-2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import "OSKCompositeSampleSource.h"

#import <Foundation/NSData.h>
#import <Foundation/NSArray.h>
#import <OmniBase/OmniBase.h>

RCS_ID("$Id: OSKCompositeSampleSource.m,v 1.11.2.1 2000/09/12 18:41:24 bungi Exp $")

@interface _OSKSubsidiarySampleSource : NSObject <OSKSubsidiarySampleSource>
{
    OSKSampleSource *sampleSource;
    OSKSampleIndex startSampleIndex;
}

- initWithSampleSource: (OSKSampleSource *) newSampleSource
      startSampleIndex: (OSKSampleIndex) newStartSampleIndex;

- (OSKSampleSource *) sampleSource;

- (OSKSampleIndex) startSampleIndex;
- (void) setStartSampleIndex: (OSKSampleIndex) newStartSampleIndex;

- (NSComparisonResult) compare: (id) SampleSource;

@end

@implementation _OSKSubsidiarySampleSource

- initWithSampleSource: (OSKSampleSource *) newSampleSource
      startSampleIndex: (OSKSampleIndex) newStartSampleIndex;
{
    sampleSource = [newSampleSource retain];
    startSampleIndex = newStartSampleIndex;

    return self;
}

- (OSKSampleSource *) sampleSource;
{
    return sampleSource;
}

- (OSKSampleIndex) startSampleIndex;
{
    return startSampleIndex;
}

- (void) setStartSampleIndex: (OSKSampleIndex) newStartSampleIndex;
{
    startSampleIndex = newStartSampleIndex;
}

- (NSComparisonResult) compare: (id) aSampleSource;
{
    OSKSampleIndex otherStart;

    otherStart = [aSampleSource startSampleIndex];
    if (otherStart > startSampleIndex)
        return NSOrderedAscending;
    else if (otherStart < startSampleIndex)
        return NSOrderedDescending;
    else
        return NSOrderedSame;
}

@end

@interface OSKCompositeSampleSource (OSKPrivateAPI)
- (void) _sampleSourcesChanged;
@end

@implementation OSKCompositeSampleSource

- initSamplingRate: (OSKSampleIndex) newSamplingRate
    bytesPerSample: (OSKSampleSize) newBytesPerSample
  samplesPerBuffer: (OSKSampleIndex) newSamplesPerBuffer;
{
    if (![super initWithSampleCount: 0
                       samplingRate: newSamplingRate
                     bytesPerSample: newBytesPerSample
                   samplesPerBuffer: newSamplesPerBuffer])
        return nil;

    subsidiarySampleSources = [[NSMutableArray alloc] init];
    activeSampleSources = [[NSMutableArray alloc] init];

    /* Since there is no method on NSData to get a zero filled buffer... */
    zeroData = [NSMutableData dataWithLength: [self bytesPerFrame] * [self framesPerBuffer]];
    zeroData = [[NSData alloc] initWithData: zeroData];

    [self _sampleSourcesChanged];
    return self;
}

- (void) dealloc;
{
    [subsidiarySampleSources release];
    [activeSampleSources release];
    [super dealloc];
}


- (void) _sampleSourcesChanged;
{
    if ([subsidiarySampleSources count]) {
        _OSKSubsidiarySampleSource *last;

        last = [subsidiarySampleSources lastObject];
        _frameCount = [last startSampleIndex] + [[last sampleSource] sampleCount];
    } else
        _frameCount = 0;

    [activeSampleSources removeAllObjects];

    /* Requested sampleIndex should never be this.  Will cause activeSampleSources to be recomputed */
    nextSampleIndex = OSK_INVALID_SAMPLE_INDEX;
    nextActivitySampleSourceIndex = 0;
    nextActivitySampleIndex = 0;
}

- (id <OSKSubsidiarySampleSource>) addSubsidiarySampleSource: (OSKSampleSource *) sampleSource
                                            startSampleIndex: (OSKSampleIndex) startSampleIndex;
{
    _OSKSubsidiarySampleSource *subsidiary;

#warning Should PRECONDITION that this sampleSource matches our parameters
    
    subsidiary = [[_OSKSubsidiarySampleSource alloc] initWithSampleSource: sampleSource
                                                         startSampleIndex: startSampleIndex];
    [subsidiarySampleSources addObject: subsidiary];
    [subsidiary release];

    /* We probably won't be doing a *ton* of these, so this shouldn't be too slow */
    [subsidiarySampleSources sortUsingSelector: @selector(compare:)];

    [self _sampleSourcesChanged];

    return subsidiary;
}

- (void) setStartSampleIndex: (OSKSampleIndex) startSampleIndex
   forSubsidiarySampleSource: (id <OSKSubsidiarySampleSource>) sampleSource;
{
    [(_OSKSubsidiarySampleSource *)sampleSource setStartSampleIndex: startSampleIndex];
    [self _sampleSourcesChanged];
}


- (void) removeSubsidiarySampleSource: (id <OSKSubsidiarySampleSource>) sampleSource;
{
    [subsidiarySampleSources removeObject: sampleSource];
    [self _sampleSourcesChanged];
}

- (NSArray *) subsidiarySampleSources;
{
    return subsidiarySampleSources;
}

/**** Stuff we have to override from our superclass ****/

static inline void _addUnsignedShortBuffers(OSKSample *dest,
                                            const OSKSample *source,
                                            unsigned int count)
{
    while (count--)
        *dest++ = *source++;
}

- (NSData *) sampleBufferStartingAtSampleIndex: (OSKSampleIndex) sampleIndex
                            beforeTimeInterval: (NSTimeInterval) beforeInterval;
{
    unsigned int activeCount;
    NSMutableData *returnData;
    _OSKSubsidiarySampleSource *subsidiary;
    OSKSampleSource *sampleSource;
    OSKSampleIndex returnSampleCount;
    OSKSample *returnBytes;
    NSData *subsidiaryData;
    const OSKSample *subsidiaryBytes;
    OSKSampleIndex subsidiaryStart, subsidiaryEnd, subsidiarySamplesToAdd;

    OBPRECONDITION(sampleIndex != OSK_INVALID_SAMPLE_INDEX);

    if (sampleIndex != nextSampleIndex) {
        /* set everything up for this index */
        unsigned int sampleSourceCount;

        [activeSampleSources removeAllObjects];

        nextActivitySampleSourceIndex = 0;
        nextActivitySampleIndex = 0;

        sampleSourceCount = [subsidiarySampleSources count];
        while (nextActivitySampleSourceIndex < sampleSourceCount) {
            _OSKSubsidiarySampleSource *subsidiary;
            unsigned int startIndex, endIndex;

            subsidiary = [subsidiarySampleSources objectAtIndex: nextActivitySampleSourceIndex];

            startIndex = [subsidiary startSampleIndex];
            if (startIndex > sampleIndex)
                /* hasn't started yet.  we've looked at enough */
                break;

            endIndex = startIndex + [[subsidiary sampleSource] sampleCount];
            if (endIndex >= sampleIndex) {
                /* hasn't already ended */
                [activeSampleSources addObject: subsidiary];
            }

            nextActivitySampleSourceIndex++;
        }

        if (nextActivitySampleSourceIndex < sampleSourceCount) {
            _OSKSubsidiarySampleSource *nextSource;

            nextSource = [subsidiarySampleSources objectAtIndex: nextActivitySampleSourceIndex];
            nextActivitySampleIndex = [nextSource startSampleIndex];
        } else
            nextActivitySampleIndex = OSK_INVALID_SAMPLE_INDEX;

        nextSampleIndex = sampleIndex;
    }

    nextSampleIndex += [self framesPerBuffer];
    activeCount = [activeSampleSources count];

    if (nextActivitySampleIndex > nextSampleIndex) {
        /* Optimize cases for zero or one active sampleSource for the duration of this buffer */
        if (activeCount == 1) {
            NSData *returnData;
            
            subsidiary = [activeSampleSources objectAtIndex: 0];
            subsidiaryStart = [subsidiary startSampleIndex];
            sampleSource = [subsidiary sampleSource];
            subsidiaryEnd = [sampleSource sampleCount] + subsidiaryStart;

            returnData = [sampleSource sampleBufferStartingAtSampleIndex: sampleIndex - subsidiaryStart
                                                      beforeTimeInterval: beforeInterval];

            if (subsidiaryEnd < nextSampleIndex)
                [activeSampleSources removeObjectAtIndex: 0];

            return returnData;
            
        } else if (!activeCount)
            return zeroData;
    }

    /* Otherwise, we have to do some mixing... start with a zeroed buffer */

    if ([self sampleCount] > nextSampleIndex)
        returnSampleCount = [self framesPerBuffer];
    else
        returnSampleCount = [self sampleCount] - sampleIndex;

    returnData = [NSMutableData dataWithLength: returnSampleCount * [self bytesPerFrame]];
    returnBytes = [returnData mutableBytes];

    /* Add in each of the currently active sample sources */
    while (activeCount--) {
        subsidiary = [activeSampleSources objectAtIndex: activeCount];
        subsidiaryStart = [subsidiary startSampleIndex];
        sampleSource = [subsidiary sampleSource];
        subsidiaryEnd = [sampleSource sampleCount] + subsidiaryStart;

        subsidiaryData = [sampleSource sampleBufferStartingAtSampleIndex: sampleIndex - subsidiaryStart
                                                      beforeTimeInterval: beforeInterval];
        subsidiaryBytes = [subsidiaryData bytes];

        if (subsidiaryEnd >= nextSampleIndex)
            subsidiarySamplesToAdd = [self framesPerBuffer];
        else {
            subsidiarySamplesToAdd = subsidiaryEnd - sampleIndex;

            /* this is the last buffer in which the sampleSource is active */
            /* since we are looping backwards, we can remove it w/o any gyrations */
            [activeSampleSources removeObject: subsidiary];
        }

        _addUnsignedShortBuffers(returnBytes, subsidiaryBytes, subsidiarySamplesToAdd);
    }

    /* Add any sampleSources that become active during this buffer */
    while (nextActivitySampleIndex < nextSampleIndex) {
        OSKSample *returnBytesOffset;

        subsidiary = [subsidiarySampleSources objectAtIndex: nextActivitySampleSourceIndex];
        nextActivitySampleSourceIndex++;

        if (nextActivitySampleSourceIndex < [subsidiarySampleSources count]) {
            _OSKSubsidiarySampleSource *next;

            next = [subsidiarySampleSources objectAtIndex: nextActivitySampleSourceIndex];
            nextActivitySampleIndex = [next startSampleIndex];
        } else
            nextActivitySampleIndex = OSK_INVALID_SAMPLE_INDEX;

        sampleSource = [subsidiary sampleSource];
        subsidiaryData = [sampleSource sampleBufferStartingAtSampleIndex: 0
                                                      beforeTimeInterval: beforeInterval];
        subsidiaryBytes = [subsidiaryData bytes];

        subsidiaryStart = [subsidiary startSampleIndex];
        subsidiaryEnd = [sampleSource sampleCount] + sampleIndex;

        returnBytesOffset = returnBytes + (subsidiaryStart - sampleIndex);

        if (subsidiaryEnd >= nextSampleIndex) {
            subsidiarySamplesToAdd = [self framesPerBuffer] - (subsidiaryStart - sampleIndex);

            /* susidiary is still active after this buffer */
            [activeSampleSources addObject: subsidiary];
        } else
            subsidiarySamplesToAdd = subsidiaryEnd - subsidiaryStart;

        _addUnsignedShortBuffers(returnBytesOffset, subsidiaryBytes, subsidiarySamplesToAdd);
    }

    return returnData;
}

@end
