// Copyright 1997-2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import <OmniBase/OmniBase.h>
#import <OmniFoundation/OmniFoundation.h>
#import "OSKResamplingFilter.h"


RCS_ID("$Id: OSKResamplingFilter.m,v 1.1.2.1 2000/09/12 18:41:25 bungi Exp $")

#define PI M_PI

@implementation OSKResamplingFilter

- initWithCutoff:(double)cyclesPerSample
     filterWidth:(int)filterSize
        channels:(int)samplesPerFrame
            gain:(double)dcGain
      sampleSize:(int)bytesPerSample
{
    double *coeff, cscale, sum, magsum, omega;
    int i;
    
    OBASSERT(filterSize > 0);
    OBASSERT(samplesPerFrame > 0);
    
    self = [super init];
    if (!self)
        return nil;

    _samplesPerFrame = samplesPerFrame;
    _bytesPerSample = bytesPerSample;
    bufferReadIndex = 0;

    coefficients = NSZoneMalloc([self zone], sizeof(*coefficients) * filterSize);

    coeff = NSZoneMalloc(NULL, sizeof(*coeff) * filterSize);
    coeff[0] = 0.5;
    magsum = 2 * coeff[0];
    sum = 2 * coeff[0];
    omega = PI * cyclesPerSample;
    for(i = 1; i < filterSize; i++) {
        coeff[i] = sin((double)i * omega) / ((double)i * omega);
        sum += 2 * coeff[i];
        magsum += 2 * fabs(coeff[i]);
    }
    scale = 0;
    cscale = dcGain / sum;

    /* compute scale to prevent possibility of overflow */
    magsum *= cscale * (1 << (8 * bytesPerSample));
    while ( ((1<<scale) * magsum) < (1L << (8 * sizeof(accumulator_t) - 2)) )
        scale ++;
    OBASSERT(scale > 0); /* if accumulator_t is too small, we lose */

    /* store scaled tap coefficients */
    for(i = 0; i < filterSize; i++)
        coefficients[i] = (int)floor(0.5 + (1<<scale) * cscale * coeff[i]);
    NSZoneFree(NULL, coeff);
    coeff = NULL;

    /* Trim any zeroes. */
    while(coefficients[filterSize-1] == 0)
        filterSize --;

    filterWidth = filterSize;

    bufferSampleCount = samplesPerFrame * 2 * (filterSize - 1);
    buffer = NSZoneMalloc([self zone], sizeof(*buffer) * bufferSampleCount);
    for(i = 0; i < bufferSampleCount; i++)
        buffer[i] = 0;

   return self;
}

- (void)addSamples:(NSData *)samples
        decimation:(unsigned)inputDecimation
{
    unsigned short channelCount, decimationCounter, tap, bytesPerSample;
    unsigned const char *sourceBuf;
    unsigned long sourceFrameCount, frameCount, decimationOffset;
    unsigned long newBufferSampleCount, sampleIndex;

    /* local copies of ivars to allow compiler optimization */
    channelCount = _samplesPerFrame;
    bytesPerSample = _bytesPerSample;
    
    sourceFrameCount = [samples length] / ( bytesPerSample * channelCount );
    OBASSERT([samples length] % (bytesPerSample * channelCount) == 0);
    sourceBuf = [samples bytes];

    frameCount = sourceFrameCount * inputDecimation;

    /* resize the buffer if needed */
    newBufferSampleCount = bufferReadIndex + channelCount * ((filterWidth-1)*2 + frameCount);
    if (newBufferSampleCount > bufferSampleCount ||
        newBufferSampleCount < (bufferSampleCount / 4)) {
        buffer = NSZoneRealloc([self zone], buffer, sizeof(*buffer) * newBufferSampleCount);
        bufferSampleCount = newBufferSampleCount;
    }

    /* zero the part of the buffer we'll use */
    for(sampleIndex = bufferReadIndex + channelCount * 2 * (filterWidth - 1);
        sampleIndex < newBufferSampleCount;
        sampleIndex ++)
        buffer[sampleIndex] = 0;

    /* convolve the (decimated) input with our coefficients, and accumulate the result */
    decimationCounter = channelCount;
    decimationOffset = bufferReadIndex;
    for(sampleIndex = 0; sampleIndex < (sourceFrameCount*channelCount); sampleIndex ++) {
        accumulator_t sample;

        switch(bytesPerSample) {
            case 1:
                sample = *(signed char *)(sourceBuf + sampleIndex);
                break;
            case 2:
                /* OSK uses an unsigned data type; however, the samples are actually signed. (OSK should probably be changed to include information on the sample format instead of just its size, like all the existing sound code. NIH much?) */
                sample = *(signed short *)(sourceBuf + sampleIndex*2);
                break;
            default:
                sample = 0;
                OBASSERT(0);
        }

        for(tap = 0; tap < filterWidth; tap ++) {
            accumulator_t tap_value = coefficients[tap] * sample;

            buffer[sampleIndex + decimationOffset + channelCount * (filterWidth - tap - 1)] += tap_value;
            buffer[sampleIndex + decimationOffset + channelCount * (filterWidth + tap - 1)] += tap_value;
        }
/*        printf("Buffer-bonkage from %ld to %ld.  (%ld): %ld  --> %ld\n",
               sampleIndex + decimationOffset + channelCount * (filterWidth - (tap-1) - 1), sampleIndex + decimationOffset + channelCount * (filterWidth + (tap-1) - 1), sample, org, buffer[sampleIndex]); */

        if (!--decimationCounter) {
            decimationCounter = channelCount;
            decimationOffset += (inputDecimation-1) * channelCount;
        }
    }

    bufferReadIndex += frameCount * channelCount;
}

- (void)flushWithZeroes
{
    NSData *zero = [[NSMutableData alloc] initWithLength:_samplesPerFrame*_bytesPerSample];
    [self addSamples:zero decimation:2 * filterWidth];
    [zero release];
}

- (NSData *)getSamplesOfSize:(OSKSampleSize)bytesPerSample
                  decimation:(unsigned)outputDecimation
{
    return [self getSamplesOfSize:bytesPerSample
                       decimation:outputDecimation
                 multiplyChannels:1];
}


- (NSData *)getSamplesOfSize:(OSKSampleSize)bytesPerSample
                  decimation:(unsigned)outputDecimation
            multiplyChannels:(unsigned)repeat
{
    NSMutableData *outputData;
    char *outputBuf;
    unsigned repeatCount;
    unsigned long sampleIndex, newBufferReadIndex;
    int shift = scale - 8 * ( bytesPerSample - _bytesPerSample );
    
    OBASSERT(outputDecimation == 1); // TODO: implement output decimation

    outputData = [[NSMutableData alloc] initWithLength: bufferReadIndex * bytesPerSample * repeat];
    outputBuf = [outputData mutableBytes];

    for (sampleIndex = 0; sampleIndex < bufferReadIndex; sampleIndex ++) {
        accumulator_t scaledSample = buffer[sampleIndex] >> shift;

//        printf("%4ld. %8.3f %5d\n", sampleIndex, ((double)(buffer[sampleIndex])) / ((double)(1 << scale)), scaledSample);

        switch(bytesPerSample) {
            case 1:
                if (scaledSample > SCHAR_MAX) { scaledSample = SCHAR_MAX; }
                else if (scaledSample < SCHAR_MIN) { scaledSample = SCHAR_MIN; }
                for(repeatCount = repeat; repeatCount; repeatCount--)
                    *((signed char *)outputBuf)++ = scaledSample;
                break;
            case 2:
                if (scaledSample > SHRT_MAX) { scaledSample = SHRT_MAX; }
                else if (scaledSample < SHRT_MIN) { scaledSample = SHRT_MIN; }
                for(repeatCount = repeat; repeatCount; repeatCount--)
                    *((signed short *)outputBuf)++ = scaledSample;
                break;
            default:
                OBASSERT(0);
        }
    }
    newBufferReadIndex = bufferReadIndex - sampleIndex;

    OBASSERT(newBufferReadIndex + 2 * (filterWidth-1) * _samplesPerFrame + sampleIndex <= bufferSampleCount);
    
    /* don't change this to memcpy without worrying about overlap */
    bcopy(buffer + sampleIndex,
          buffer,
          (newBufferReadIndex + 2 * (filterWidth-1) * _samplesPerFrame) * bytesPerSample);

    bufferReadIndex = newBufferReadIndex;

    return [outputData autorelease];
}

- (int)groupDelay
{
    return filterWidth - 1;
}

- (unsigned char)samplesPerFrame
{
    return _samplesPerFrame;
}
    
@end
