// Copyright 1997-2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import "OSKResamplingStream.h"
#import "OSKSampleBuffer.h"
#import "OSKResamplingFilter.h"

#import <OmniBase/OmniBase.h>

RCS_ID("$Id: OSKResamplingStream.m,v 1.2.2.1 2000/09/12 18:41:26 bungi Exp $")

@interface OSKResamplingStream (Private)
- (void)_setupFilter;
@end

@implementation OSKResamplingStream

- initSampleRate: (OSKSamplingRate) rate
  bytesPerSample: (int)bps
 samplesPerFrame: (int)spf
          source: (OSKSampleStream *)source
   initialBuffer: (OSKSampleBuffer *)buffer
{
    self = [super init];
    if (!self)
        return nil;

    filter = nil;
    nextBuffer = [buffer retain];
    bufferSource = [source retain];

    _samplingRate = rate;
    _numberOfChannelsIn = 0;
    _bytesPerSample = bps;
    _numberOfChannelsOut = spf;

    if (buffer)
        [self _setupFilter];

    return self;
}

- (void)dealloc
{
    [nextBuffer release];
    [filter release];
    [bufferSource release];
    [super dealloc];
}
    
- (void)_setupFilter
{
    OSKSamplingRate sourceRate, filterRate;
    
    OBASSERT(filter == nil);
    OBASSERT(_numberOfChannelsIn == 0);
    OBASSERT(nextBuffer != nil);

    _numberOfChannelsIn = [nextBuffer numberOfChannels];
    sourceRate = [nextBuffer samplingRate];

    /* Right now this code will only do upsampling by integer multiples, although OSKResamplingFilter can do any small rational rate conversion. */
    OBASSERT(_samplingRate % sourceRate == 0);
    filterRate = _samplingRate;

    inputDecimation = filterRate / sourceRate;
    
    filter = [[OSKResamplingFilter allocWithZone:[self zone]]
        initWithCutoff: (double)sourceRate / (double)filterRate
           filterWidth: 30 /* TODO: Dynamically determine the filter width */
              channels: _numberOfChannelsIn
                  gain: inputDecimation
            sampleSize: [nextBuffer bytesPerSample]];

    groupDelayTicks = OSKSamplesToTicks([filter groupDelay], filterRate);

    OBASSERT(_numberOfChannelsOut % _numberOfChannelsIn == 0);
}

- (BOOL)isAtEnd
{
    if (!nextBuffer &&
        (!bufferSource || [bufferSource isAtEnd]))
        return YES;
    return NO;
}

- nextSampleBufferByInterval: (NSTimeInterval)deadline
         minimumStartingTick: (OSKTick)minimumTick
{
    NSData *filteredData;
    OSKSampleBuffer *result;
    OSKTick resultTick;
    
    if (!nextBuffer) {
        nextBuffer = [[bufferSource nextSampleBufferByInterval:deadline minimumStartingTick:minimumTick] retain];
        if (!nextBuffer)
            return nil;
    }

    if (!filter)
        [self _setupFilter];

    // TODO: Assert that nextBuffer's sample size, sampling rate, number of channels, etc., are what we expect.

    [filter addSamples:[nextBuffer sampleData]
            decimation:inputDecimation];

    resultTick = [nextBuffer startingTick];
    if (resultTick != OSK_INVALID_TICK) {
        if (resultTick > groupDelayTicks)
            resultTick -= groupDelayTicks;
        else
            resultTick = 0;
    }
    
    [nextBuffer release];
    nextBuffer = nil;

    if ([bufferSource isAtEnd])
        [filter flushWithZeroes];

    filteredData = [filter getSamplesOfSize:_bytesPerSample
                                 decimation:1
                           multiplyChannels:_numberOfChannelsOut / _numberOfChannelsIn];

    // NSLog(@"resampled %u bytes to %u", ol, [filteredData length]);

    result = [[OSKSampleBuffer alloc] initWithSampleData: filteredData
                                            samplingRate: _samplingRate
                                        numberOfChannels: _numberOfChannelsOut
                                          bytesPerSample: _bytesPerSample
                                            startingTick: resultTick];

    return [result autorelease];
}
        
@end
