// Copyright 1997-2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import "OSKSNDFileSampleSource.h"

#import <Foundation/NSPathUtilities.h>
#import <Foundation/NSData.h>
#import <Foundation/NSException.h>
#import <Foundation/NSString.h>

#import "SoundKit.h"


#import "OSKSequentialFileSampleSource.h"
#import "OSKDataSampleSource.h"

#import <OmniBase/OmniBase.h>
#import <OmniBase/system.h>

RCS_ID("$Id: OSKSNDFileSampleSource.m,v 1.7.2.2 2000/09/12 19:05:42 bungi Exp $")

static NSString *OSKBadSNDFileException = @"OSKBadSNDFileException";
static NSString *OSKUnsupportedSNDFileFormatException = @"OSKUnsupportedSNDFileFormatException";
 
@implementation OSKSNDFileSampleSource

+ sampleSourceWithSNDFileAtPath: (NSString *) aPath
                framesPerBuffer: (OSKSampleIndex) newFramesPerBuffer
             sequentialReadOnly: (BOOL) sequential;
{
    int            fileDescriptor;
    SNDSoundStruct soundStruct;
    OSKSampleIndex dataSize; // .snd files only have a 'unsigned long', not a full 'unsigned long long' though
    OSKSampleSize  bytesPerFrame;
    OSKSampleSize  bytesPerSample;


    OBPRECONDITION(aPath);

    if (![[aPath pathExtension] isEqualToString: @"snd"])
        NSLog(@"Suspicious filename for a SND file: %@", aPath);

    if ((fileDescriptor = open([aPath cString], O_RDONLY)) < 0)
        [NSException raise: OSKBadSNDFileException format: @"Cannot open %@", aPath];

    if (read(fileDescriptor, &soundStruct, sizeof(soundStruct)) != sizeof(soundStruct))
        [NSException raise: OSKBadSNDFileException format: @"Couldn't read sound header"];

    /* Verify that this is a SND file and grab all the info we'll need */
    /* TODO: Should be able to read files in non-native byte order; see NeXT's implementation for details */
    if (soundStruct.magic != SND_MAGIC)
        [NSException raise: OSKBadSNDFileException format: @"Bad magic number in sound file %@", aPath];
    switch (soundStruct.dataFormat) {
        case SND_FORMAT_LINEAR_8:
            bytesPerSample = 1; break;
        case SND_FORMAT_LINEAR_16:
            bytesPerSample = 2; break;
        case SND_FORMAT_LINEAR_24:
            bytesPerSample = 3; break;
        default:
            [NSException raise: OSKUnsupportedSNDFileFormatException
                        format: @"Unsupport data format.  Must currently be 8, 16, or 24 bit linear."];
            bytesPerSample = 0;  /* make the compiler happy */
    }

    if (!(dataSize = (OSKSampleIndex)soundStruct.dataSize)) {
        /* This field isn't always filled in for some moronic reason */
        /* (eg, if OmniSoundKit wrote the file, it won't be filled) */
        struct stat statBuf;

        if (fstat(fileDescriptor, &statBuf) != 0)
            [NSException raise: OSKBadSNDFileException format: @"Couldn't determine sound file length"];
        dataSize = (OSKSampleIndex)(statBuf.st_size) - (OSKSampleIndex)soundStruct.dataLocation;
    }

    /* Make sure the fileDescriptor is positioned at the beginning of the sound data */
    if (soundStruct.dataLocation != sizeof(soundStruct)) {
        if (lseek(fileDescriptor, L_SET, soundStruct.dataLocation) < 0)
            [NSException raise: OSKBadSNDFileException format: @"Couldn't seek to beginning of sound data"];
    }

    bytesPerFrame = bytesPerSample * soundStruct.channelCount;

    if (sequential) {
        return [[[OSKSequentialFileSampleSource alloc]
                initWithFileDescriptor: fileDescriptor
                            frameCount: dataSize / bytesPerFrame
                          samplingRate: OSKValidateSamplingRate(soundStruct.samplingRate)
                              channels: soundStruct.channelCount
                        bytesPerSample: bytesPerSample
                       framesPerBuffer: newFramesPerBuffer] autorelease];
    } else {
        NSData *fileData, *sampleData;
        NSRange sampleRange;

#warning FIXME: This is suboptimal due to Foundation badness
/* Foundation should have API for mapping only a portion of a file.  Otherwise, we will be doing lots of non-page aligned operations */

        fileData = [[NSData alloc] initWithContentsOfMappedFile: aPath];
        sampleRange.location = soundStruct.dataLocation;
        sampleRange.length = dataSize;

        sampleData = [fileData subdataWithRange: sampleRange];
        [fileData release];

        return [[[OSKDataSampleSource alloc] initWithSampleData: sampleData
                                                   samplingRate: soundStruct.samplingRate
                                                       channels: soundStruct.channelCount
                                                 bytesPerSample: bytesPerSample
                                                framesPerBuffer: newFramesPerBuffer] autorelease];
    }
}

@end
