// Copyright 1997-2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import <Foundation/NSData.h>

#import "OSKSampleSource.h"

#import <OmniBase/OmniBase.h>

RCS_ID("$Id: OSKSampleSource.m,v 1.12.2.1 2000/09/12 18:41:27 bungi Exp $")

@implementation OSKSampleSource

/*"
Initializes the receiver.  The sample count indicates the number of distinct points in time for which the receiver will provide data.  Since all OSKSampleSources provide stereo data, this means that each sample will provide two data points at each point in time.

This is a legacy method; new code should use -initWithFrameCount:samplingRate:channels:framesPerBuffer:.
"*/
- initWithSampleCount: (OSKSampleIndex) newSampleCount
         samplingRate: (OSKSamplingRate) newSamplingRate
       bytesPerSample: (OSKSampleSize) newBytesPerSample
     samplesPerBuffer: (OSKSampleIndex) newSamplesPerBuffer;
{
    OBPRECONDITION(newSamplingRate == OSK44100SamplingRate);
    OBPRECONDITION(newBytesPerSample == 4);

    return [self initWithFrameCount:newSampleCount
                       samplingRate:newSamplingRate
                           channels:2
                     bytesPerSample:2
                    framesPerBuffer:newSamplesPerBuffer];
}

/*"
This is the designatied initializer for OSKSampleSource. The frame count indicates the number of distinct points in time for which the receiver will provide data, or OSK_INVALID_SAMPLE_INDEX if this is unknown. The sample rate indicates the frequency at which the receiver can provide data; the channel count indicates the number of samples taken at each sampling instant (e.g. 2 for stereo sound), and bytesPerSample indicates the precision of each sample. Since OSKSampleSource only handles linear encoded data, the sample size determines the sample format. framesPerBuffer indicates the number of frames that the receiver will return per call to -sampleBufferStartingAtSampleIndex:beforeTimeInterval:.
"*/
- initWithFrameCount: (OSKSampleIndex) newSampleCount
        samplingRate: (OSKSamplingRate) newSamplingRate
            channels: (int) newSamplesPerFrame
      bytesPerSample: (OSKSampleSize) newBytesPerSample
     framesPerBuffer: (OSKSampleIndex) newSamplesPerBuffer;
{
    [super init];

    OBPRECONDITION(newSamplesPerBuffer > 0);
    OBPRECONDITION(newSamplesPerFrame > 0);
    OBPRECONDITION(newBytesPerSample > 0);
    OBPRECONDITION(newSamplingRate > 0);

    _frameCount      = newSampleCount;
    _samplingRate    = newSamplingRate;
    _framesPerBuffer = newSamplesPerBuffer;
    _bytesPerSample  = newBytesPerSample;
    _samplesPerFrame = newSamplesPerFrame;
    
    return self;
}


/*"
Returns the number of sampling instants covered by the receiver. This is equal to the total number of frames, which will not be equal to the number of samples if the channel count is not 1.
"*/
- (OSKSampleIndex) sampleCount;
{
    return _frameCount;
}

/*"
Returns the number of sampling instants covered by the receiver. This is equal to the total number of frames, which will not be equal to the number of samples if the channel count is not 1.
"*/
- (OSKSampleIndex) frameCount;
{
    return _frameCount;
}

/*"
Returns the number of samples that should be played per second per channel.  Currently, this must be 44100.
"*/
- (OSKSamplingRate) samplingRate;
{
    return _samplingRate;
}

/*"
Returns the number of bytes of data needed to represent a complete sample.  Currently this must be 4 (indicating 2 16 bit channels).
"*/
- (OSKSampleSize) bytesPerFrame;
{
    return _bytesPerSample * _samplesPerFrame;
}

/*"
Returns the number of bytes per sample. Typically this is 2, for a 16-bit liear sample format.
"*/
- (OSKSampleSize) bytesPerSample
{
    return _bytesPerSample;
}

/*"
Returns the number of samples taken at each sampling instant, e.g. 2 for stereo sound.
"*/
- (int) samplesPerFrame
{
    return _samplesPerFrame;
}

/*"
Returns the number of sample frames (each bytesPerFrame in size) that should be returned in each streaming buffer returned by -sampleBufferStartingAtSampleIndex:beforeTimeInterval:.  Note that the receiver can opt to return fewer frames if need be.
"*/
- (OSKSampleIndex) framesPerBuffer;
{
    return _framesPerBuffer;
}

/*"
Returns the number of seconds of audio that can be produced by playing this sample source.  This is derived from the other parameters on OSKSampleBuffer.
"*/
- (NSTimeInterval) sampleSourceDuration;
{
    /* There could be some serious rounding error in this with really big sample counts */
    return OSKSamplesToSeconds([self sampleCount], [self samplingRate]);
}

//
// Methods that must be subclassed
//

/*"
Returns the portion of the sample data for the receiver, starting from the given sample index and containing at most -samplesPerBuffer samples.  The actual amount of contained data may be less (if for example you are requesting the last buffer).

The receiver should make every attempt to return before the given time interval has passed.  If only a small amount of data has been obtained by that interval, it can be returned.  If no data has been obtained, nil should be returned.
"*/
- (NSData *) sampleBufferStartingAtSampleIndex: (OSKSampleIndex) sampleIndex
                            beforeTimeInterval: (NSTimeInterval) beforeInterval;
{
    OBRequestConcreteImplementation([self class], _cmd);
    return nil;
}

@end
