// Copyright 1997-2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import "OSKSequentialFileSampleSource.h"

#import <Foundation/Foundation.h>

#import <OmniFoundation/NSThread-OFExtensions.h>
#import <OmniFoundation/OFQueue.h>
#import <OmniBase/OmniBase.h>
#import <OmniBase/system.h>

static NSString *OSKQueueException             = @"OSKQueueException";
static NSString *OSKFileException              = @"OSKFileException";
static NSString *OSKNonSequentialReadException = @"OSKNonSequentialReadException";

static BOOL OSKSequentialFileSampleSourceDebug = NO;

RCS_ID("$Id: OSKSequentialFileSampleSource.m,v 1.9.2.1 2000/09/12 18:41:28 bungi Exp $")

@interface OSKSequentialFileSampleSource (Private)
- (void) _getNextSegment;
@end

static OSKSampleIndex defaultBuffersPerSegment = (OSKSampleIndex)64;
static OSKSampleIndex defaultSegmentQueueSize  = (OSKSampleIndex)32;

@implementation OSKSequentialFileSampleSource

+ (void) setDefaultBuffersPerSegment: (OSKSampleIndex) newBuffersPerSegment;
{
    OBPRECONDITION(newBuffersPerSegment);

    defaultBuffersPerSegment = newBuffersPerSegment;
}

+ (void) setDefaultSegmentQueueSize: (OSKSampleIndex) newSegmentQueueSize;
{
    OBPRECONDITION(newSegmentQueueSize);

    defaultSegmentQueueSize = newSegmentQueueSize;
}

- initWithFileDescriptor: (int) fileDescriptor
             sampleCount: (OSKSampleIndex) newSampleCount
            samplingRate: (OSKSamplingRate) newSamplingRate
          bytesPerSample: (OSKSampleSize) newBytesPerSample
        samplesPerBuffer: (OSKSampleIndex) newSamplesPerBuffer;
{
    return [self initWithFileDescriptor: fileDescriptor
                             frameCount: newSampleCount
                           samplingRate: newSamplingRate
                               channels: 2
                         bytesPerSample: newBytesPerSample/2
                        framesPerBuffer: newSamplesPerBuffer];
}

- initWithFileDescriptor: (int) fileDescriptor
              frameCount: (OSKSampleIndex) newFrameCount
            samplingRate: (OSKSamplingRate) newSamplingRate
                channels: (int) newSamplesPerFrame
          bytesPerSample: (OSKSampleSize) newBytesPerSample
         framesPerBuffer: (OSKSampleIndex) newFramesPerBuffer;
{
    OBPRECONDITION(fileDescriptor > 0);

    if (![self initWithFrameCount: newFrameCount
                     samplingRate: newSamplingRate
                         channels: newSamplesPerFrame
                   bytesPerSample: newBytesPerSample
                  framesPerBuffer: newFramesPerBuffer])
        return nil;

    fd = fileDescriptor;

    buffersPerSegment = defaultBuffersPerSegment;
    framesPerSegment = buffersPerSegment * _framesPerBuffer;

    segmentSize = framesPerSegment * _samplesPerFrame * _bytesPerSample;
    totalSize = _frameCount * _samplesPerFrame * _bytesPerSample;

    currentSegmentStartFrameIndex = 0;
    currentSegmentEndFrameIndex = 0;

    segmentReaderStarted = NO;
    currentSegmentData = nil;

    segmentQueue = [[OFQueue alloc] initWithCount: defaultSegmentQueueSize];
    segmentQueueLowWaterLock = [[NSLock alloc] init];
    segmentQueueHasFilled = NO;

    return self;
}

- (void) _readSegments;
{
    NSAutoreleasePool *pool;
    NSMutableData     *readBuffer;
    OSKSampleIndex     readFrameIndex = 0;

    while (readFrameIndex < _frameCount) {
        OSKSampleIndex nextFrameIndex;
        OSKSampleIndex readFrameCount;
        OSKSampleIndex readLength;

        pool = [[NSAutoreleasePool alloc] init];
        
        nextFrameIndex = readFrameIndex + framesPerSegment;
        if (nextFrameIndex > _frameCount)
            nextFrameIndex = _frameCount;

        readFrameCount = nextFrameIndex - readFrameIndex;

        readLength = readFrameCount * _samplesPerFrame * _bytesPerSample;
        readBuffer = [[NSMutableData alloc] initWithLength: readLength];
        if (OSKSequentialFileSampleSourceDebug) {
            off_t offset;

            offset = lseek(fd, (off_t)0, L_INCR);
            NSLog(@"reading 0x%08x bytes from offset [0x%08x:%08x]", (unsigned int)readLength, (unsigned int)(offset >> 32), (unsigned int)offset);
        }
        if (read(fd, [readBuffer mutableBytes], (off_t)readLength) != readLength) {
            [segmentQueue close];
            [readBuffer release];
            [NSException raise: OSKFileException format: @"Failed to read %d bytes of sound file", segmentSize];
        }
        if (0 && OSKSequentialFileSampleSourceDebug) {
            NSLog(@"readBuffer = %@", readBuffer);
        }
        
        while (![segmentQueue enqueue: readBuffer shouldWait: segmentQueueHasFilled]) {
            /* We are full -- let the other thread know */
            segmentQueueHasFilled = YES;
            [segmentQueueLowWaterLock unlock];
        }
        [readBuffer release];

        readFrameIndex += readFrameCount;

        [pool release];
    }

    [segmentQueue close];

    if (!segmentQueueHasFilled) {
        /* It's as full as it's going to get */
        segmentQueueHasFilled = YES;
        [segmentQueueLowWaterLock unlock];
    }

    if (OSKSequentialFileSampleSourceDebug)
        NSLog(@"Done reading source file");
}

- (void) _getNextSegment;
{
    OBPRECONDITION(currentSegmentEndFrameIndex < _frameCount);


    if (!segmentReaderStarted) {
        segmentReaderStarted = YES;
        [segmentQueueLowWaterLock lock];

        [NSThread detachNewThreadSelector: @selector(_readSegments) toTarget: self withObject: nil];

        /* Now, block until the reader has filled up the queue at least once */
        if (OSKSequentialFileSampleSourceDebug)
            NSLog(@"Filling segment queue");
        [segmentQueueLowWaterLock lock];
        if (OSKSequentialFileSampleSourceDebug)
            NSLog(@"Segment queue full");
        [segmentQueueLowWaterLock unlock];
    }
    
    currentSegmentStartFrameIndex = currentSegmentEndFrameIndex;
    currentSegmentEndFrameIndex += framesPerSegment;
    if (currentSegmentEndFrameIndex > _frameCount)
        currentSegmentEndFrameIndex = _frameCount;

    // Get rid of the old segment
    [currentSegmentData release];
    currentSegmentData = nil;
    
    if (!(currentSegmentData = [segmentQueue dequeueShouldWait: YES]))
        [NSException raise: OSKQueueException format: @"Failed to dequeue the next segment in the sound stream"];

    if ([segmentQueue maxCount] - [segmentQueue count] > 2 && ![segmentQueue isClosed]) {
        [[NSThread currentThread] yield];
    }

    /* We might want to turn this into an immutable data here to avoid having to do vm_copy when making subdatas */
    [currentSegmentData retain];
}

- (NSData *) sampleBufferStartingAtSampleIndex: (OSKSampleIndex) sampleIndex
                            beforeTimeInterval: (NSTimeInterval) beforeInterval;
{
    NSRange subdataRange;
    OSKSampleIndex subdataFrameCount;

    if (sampleIndex < currentSegmentStartFrameIndex)
        [NSException raise: OSKNonSequentialReadException format: @"%@ only allows sequential reads", self];
    if (sampleIndex >= currentSegmentEndFrameIndex) {
        if (sampleIndex == currentSegmentEndFrameIndex) {
            if (OSKSequentialFileSampleSourceDebug)
                NSLog(@"Getting segment at sampleIndex 0x%08x:%08x", (unsigned int)(sampleIndex >> 32), (unsigned int)(sampleIndex & 0xffffffff));
            [self _getNextSegment];
        } else
            [NSException raise: OSKNonSequentialReadException format: @"%@ only allows sequential reads", self];
    }

    /* return the appropriate subdata from the current segment data */
    subdataFrameCount = MIN(_framesPerBuffer, currentSegmentEndFrameIndex - sampleIndex);
    subdataRange.location = (sampleIndex - currentSegmentStartFrameIndex) * _samplesPerFrame * _bytesPerSample;
    subdataRange.length = subdataFrameCount * _samplesPerFrame * _bytesPerSample;

    if (OSKSequentialFileSampleSourceDebug) {
        NSLog(@"Returning subdata spanning 0x%08x - 0x%08x",
              (unsigned int)subdataRange.location,
              (unsigned int)(subdataRange.location + subdataRange.length - 1));
    }

    return [currentSegmentData subdataWithRange: subdataRange];
}

@end
