// Copyright 1997-2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import "OSKSoundOutSampleDestination.h"

#import "OSKSampleStream.h"
#import "OSKSampleBuffer.h"
#import "OSKResamplingStream.h"

#import <Foundation/Foundation.h>
#import "SoundKit.h"

#import <OmniFoundation/OmniFoundation.h>
#import <OmniBase/OmniBase.h>
#import <mach/mach_error.h>

RCS_ID("$Id: OSKSoundOutSampleDestination.m,v 1.48.2.2 2000/09/12 19:05:42 bungi Exp $")

static void play_aborted(void *userData, int tag);
static void play_started(void *userData, int tag);
static void play_completed(void *userData, int tag);
static void play_overflow(void *userData, int tag);

static unsigned int OSKSoundDriverDebugLevel = 0;
static BOOL OSKSoundDriverExitRatherThanResettingNextTick = NO;


// We use a monotonically increasing series of unsigned ints as tags for buffer
// records.  If were were to enqueue 8k (one or two pages of data, depending
// up the arch) on each stream operation, playing 16 bit stereo at 44.1kHz,
// a 32-bit unsigned int gives enough tags for 6.3 years of continuous audio.
typedef struct _OSKBufferRecord {
    // If tag is non-zero, this group of information is valid
    unsigned int     tag;
    NSTimeInterval   completionInterval;
    NSTimeInterval   desiredStartInterval;
    NSTimeInterval   estimatedStartInterval;
    NSTimeInterval   duration;
    OSKSampleBuffer *sampleBuffer;
    BOOL             isSilence;

    // If hasStarted is YES, this group of information is valid
    BOOL             hasStarted;
    NSTimeInterval   actualStartInterval;
    
    // These are stats for average computations
    NSTimeInterval   lastStartOffset;
} OSKBufferRecord;


static NSString *OSKSoundOutException = @"OSKSoundOutException";
static OSKSoundOutSampleDestination *sharedSoundOutSampleDestination = nil;
static OSKSampleIndex  soundLowWater = (OSKSampleIndex)0;
static OSKSampleIndex  soundHighWater = (OSKSampleIndex)0;
static NSPort         *replyPort = nil;

static boolean_t preempt, deallocateWhenDone, msgStarted, msgCompleted;
static boolean_t msgResumed, msgUnderrun, msgAborted, msgPaused;

static OSKSampleBuffer    *zeroSampleBuffer = nil;
static OFFastMutableData  *underflowZeroData = nil;

@interface OSKSoundOutSampleDestination (PrivateAPI)
- (void) _locked_releaseRecord: (OSKBufferRecord *) bufferRecord;
- (BOOL) _locked_waitForAvailableQueuePosition;
- (void) _play;
- (void) _locked_enqueueSampleBuffer: (OSKSampleBuffer *) sampleBuffer
              estimatedStartInterval: (NSTimeInterval) estimatedStartInterval
                           isSilence: (BOOL) isSilence;
- (BOOL) _locked_containsOnlySilence: (BOOL) andIsFull;
- (NSTimeInterval) _locked_minimumSynchronizationOffset;
+ (void) _readReplyMessages;
- (void) _locked_getNewSampleStream;
- (void) _locked_setNewSampleStream:(OSKSampleStream *)newStream;
- (NSTimeInterval) _locked_nextBufferStartTimeInterval: (BOOL) fromNow
                                           bufferCount: (unsigned int *) bufferCount;
- (void) _playStarted: (int) tag;
- (void) _playCompleted: (int) tag;
@end


static BOOL UseFixedPriorityMode = NO;

@implementation OSKSoundOutSampleDestination

- _init;
{
    OFUserDefaults     *defaults;
    OFFastMutableData  *zeroSampleData;

    UseFixedPriorityMode = [[OFUserDefaults sharedUserDefaults] boolForKey: @"UseFixedPriorityMode"];
    NSLog(@"UseFixedPriorityMode = %d", UseFixedPriorityMode);
    
    /* Set the snddriver_stream_start_writing options we'll use */
    preempt            = 0;
    deallocateWhenDone = 0;
    msgStarted         = 1;
    msgCompleted       = 1;
    msgAborted         = 1;
    msgPaused          = 0;
    msgResumed         = 0;
    msgUnderrun        = 1;


    defaults = [OFUserDefaults sharedUserDefaults];
    OSKSoundDriverDebugLevel = [defaults integerForKey:@"OSKSoundDriverDebugLevel"];
    if (OSKSoundDriverDebugLevel != 0)
        NSLog(@"OSKSoundDriverDebugLevel = %d", OSKSoundDriverDebugLevel);
    OSKSoundDriverExitRatherThanResettingNextTick = [defaults boolForKey:@"OSKSoundDriverExitRatherThanResettingNextTick"];

    _numberOfBuffers = [defaults integerForKey:@"OSKSoundDriverNumberOfBuffers"];

    _mutex = NSZoneMalloc(NULL, sizeof(*_mutex));
    mutex_init(_mutex);

    _condition = NSZoneMalloc(NULL, sizeof(*_condition));
    condition_init(_condition);

    _locked_devicePort = nil;
    _locked_ownerPort  = nil;
    _locked_streamPort = nil;
    _locked_playState  = OSKStoppedPlayState;
    _locked_sampleStream = nil;
    _locked_recordQueue = NULL;
    _locked_nextFinishingQueueEntry = 0;
    _locked_nextTag = 1;
    
    _pauseLock = [[NSLock alloc] init];
    _locked_pauseCount = 0;

    _doneLock = [[NSLock alloc] init];

    // This is used for flushing out any pending samples when the stream ends.
    zeroSampleData = [OFFastMutableData newFastMutableDataWithLength: NSPageSize()/4];
    [underflowZeroData fillWithZeros];

    zeroSampleBuffer = [[OSKSampleBuffer alloc] initWithSampleData: zeroSampleData
                                                      samplingRate: OSK44100SamplingRate
                                                  numberOfChannels: 2
                                                    bytesPerSample: sizeof(OSKSample)
                                                      startingTick: OSK_INVALID_TICK];

    // Our minimum underflow silence will be 0.5 seconds.  This has the side effect
    // that we won't constantly be enqueuing very short buffers when nothing is being
    // played in OmniAmp
    underflowZeroData = [OFFastMutableData newFastMutableDataWithLength: 0.5 * OSK44100SamplingRate * 2 * sizeof(OSKSample)];
    [underflowZeroData fillWithZeros];
    
    replyPort = [[NSPort alloc] init];
    [NSThread detachNewThreadSelector: @selector(_readReplyMessages) toTarget: isa withObject: nil];

    return self;
}

+ (void) initialize;
{
    [super initialize];

    if (soundLowWater)
        return;

    [self setSoundLowWaterBuffers: 32];
    [self setSoundHighWaterBuffers: 128];
}

+ (void) setSoundLowWaterBuffers: (OSKSampleIndex) newLowWaterBuffers;
{
    OBPRECONDITION(newLowWaterBuffers);

    soundLowWater = newLowWaterBuffers * (NSPageSize()/2);
}

+ (void) setSoundHighWaterBuffers: (OSKSampleIndex) newHighWaterBuffers;
{
    OBPRECONDITION(newHighWaterBuffers);

    soundHighWater = newHighWaterBuffers * (NSPageSize()/2);
}

+ (OSKSoundOutSampleDestination *) sharedSoundOutSampleDestination;
{
    if (!sharedSoundOutSampleDestination)
        sharedSoundOutSampleDestination = [[self alloc] _init];
    return sharedSoundOutSampleDestination;
}

- init;
{
    OBRejectUnusedImplementation([self class], _cmd);
    return nil;
}

- (void) setDelegate: (id) delegate;
{
    _delegate = delegate;
}

- (id) delegate;
{
    return _delegate;
}

- (void) setNumberOfBuffers: (unsigned int) bufferCount;
{
    mutex_lock(_mutex);

    if (_locked_playState != OSKStoppedPlayState) {
        mutex_unlock(_mutex);
        [NSException raise: @"IllegalCommand"
                    format: @"Cannot set the number of buffers when not stopped."];
    } else if (bufferCount < 2) {
        mutex_unlock(_mutex);
        [NSException raise: @"IllegalCommand"
                    format: @"Must have at least two buffers."];
    }

    _numberOfBuffers = bufferCount;

    mutex_unlock(_mutex);
}

- (unsigned int) numberOfBuffers;
{
    return _numberOfBuffers;
}

/*"
This configures the receiver to play samples from the given stream.  The baseTimeInterval indicates the time at which a sample buffer with a -startingTick of zero should start playing.  These must be set together, primarily to help callers avoid forgetting to set the base time interval.
 "*/
- (void) setSampleStream: (OSKSampleStream *) newSampleStream
        baseTimeInterval: (NSTimeInterval) baseTimeInterval;
{
    [self stop];

    mutex_lock(_mutex);
    [self _locked_setNewSampleStream:newSampleStream];
    _baseTimeInterval = baseTimeInterval;
    if (OSKSoundDriverDebugLevel > 2)
        NSLog(@"OSKSoundOutSampleDestination: _locked_sampleStream = %@, _baseTimeInterval = %f",
              _locked_sampleStream, _baseTimeInterval);

    mutex_unlock(_mutex);
}

/*"
Returns the time interval at which a sample buffer with a -startingTick of zero should start playing.
"*/
- (NSTimeInterval) baseTimeInterval;
{
    return _baseTimeInterval;
}

/*"
Returns the average time interval that the start of each recent packet differed from its estimated start time.
"*/
- (NSTimeInterval) averageSynchronizationOffset;
{
    NSTimeInterval minimumSynchronizationOffset;

    mutex_lock(_mutex);
    minimumSynchronizationOffset = [self _locked_minimumSynchronizationOffset];
    mutex_unlock(_mutex);

    return minimumSynchronizationOffset;
}

/*"
Returns the estimated duration of the currently enqueued samples.  This includes the full time of the unstarted buffers and a prorated portion of the playing buffer based on when it started, its duration and the current time.
"*/
- (NSTimeInterval) enqueuedSampleDuration;
{
    NSTimeInterval duration;
    unsigned int   bufferCount;
    
    mutex_lock(_mutex);
    duration = [self _locked_nextBufferStartTimeInterval: YES bufferCount: &bufferCount];
    mutex_unlock(_mutex);

    return duration;
}

- (unsigned int) insertSamplesAdjustmentCount;
{
    return _insertSamplesAdjustmentCount;
}

- (unsigned int) removeSamplesAdjustmentCount;
{
    return _removeSamplesAdjustmentCount;
}

/*"
Returns the currently playing sample stream, or nil if no sample stream is set.
"*/
- (OSKSampleStream *) sampleStream;
{
    OSKSampleStream *sampleStream;

    mutex_lock(_mutex);
    sampleStream = [[_locked_sampleStream retain] autorelease];
    mutex_unlock(_mutex);

    return sampleStream;
}

#define OSK_SOUND_PRIORITY   (20)
#define OSK_SOUND_PREEMPT    (1)
#define OSK_SOUND_UNUSED     (0)


static int _SNDNegotiationFun(void *arg, SNDNegotiationData *data)
{
    fprintf(stderr, "_SNDNegotiationFun called.\n");
    return 0;
}

- (void) start;
{
    NSAutoreleasePool *pool;
    kern_return_t rc;
    int           protocol = 0, streamType;
    OSKSampleSize bytesPerSample;
    port_t        devicePort, ownerPort, streamPort;
    SNDError      sndError;

    pool = [[NSAutoreleasePool alloc] init];
    
    if (OSKSoundDriverDebugLevel > 0)
        NSLog(@"OSKSoundOutSampleDestination: -start called");

    mutex_lock(_mutex);

    // Are we already started?
    if (_locked_playState != OSKStoppedPlayState) {
        mutex_unlock(_mutex);
        [pool release];
        return;
    }

    // We'll unlock this when we are done so that anyone waiting in
    // -finish will get unblocked
    [_doneLock lock];

    
    if (!_locked_sampleStream || [_locked_sampleStream isAtEnd]) {
        [self _locked_getNewSampleStream];

        if (!_locked_sampleStream || [_locked_sampleStream isAtEnd]) {
            [_doneLock unlock];
            mutex_unlock(_mutex);
            [pool release];
            [NSException raise: @"IllegalCommand"
                        format: @"Cannot start a sound out sample destination w/o a sampleSource"];
        }
    }

    if (OSKSoundDriverDebugLevel > 1)
        NSLog(@"OSKSoundOutSampleDestination: obtained first sample stream of %@", _locked_sampleStream);

    // For some inexplicable reason (well, it might have to do with some negotiation stuff), after we
    // send -stop, this will fail the first time with SND_ERR_CANNOT_ACCESS.  It will work the second time.
    devicePort = ownerPort = PORT_NULL;
    sndError = SNDAcquire(SND_ACCESS_OUT, OSK_SOUND_PRIORITY, OSK_SOUND_PREEMPT, OSK_SOUND_UNUSED,
                          _SNDNegotiationFun, OSK_SOUND_UNUSED, &devicePort, &ownerPort);
    if (sndError != SND_ERR_NONE) {
        [_doneLock unlock];
        mutex_unlock(_mutex);
        [pool release];
        [NSException raise: OSKSoundOutException
                    format: @"SNDAcquire failed, sndError = %d, %s", sndError, SNDSoundError(sndError)];
    }

    _locked_recordQueue = NSZoneCalloc(NULL, _numberOfBuffers, sizeof(OSKBufferRecord));
    _locked_nextFinishingQueueEntry = 0;

    _locked_devicePort = [[NSPort alloc] initWithMachPort: devicePort];
    _locked_ownerPort  = [[NSPort alloc] initWithMachPort: ownerPort];
    

#if 1
    // It would be nice if we could change the sampling rate on the sound-out mid-stream.
    // This is probably not possible w/o stopping the stream though (at least on the vast majority of hardware).
    streamType     = SNDDRIVER_STREAM_TO_SNDOUT_44;
    bytesPerSample = sizeof(OSKSample);
#else
    streamType = ( [_locked_sampleSource samplingRate] == OSK44100SamplingRate ) ?
        SNDDRIVER_STREAM_TO_SNDOUT_44 :
        SNDDRIVER_STREAM_TO_SNDOUT_22;
    bytesPerSample   = [_locked_sampleSource bytesPerSample];
#endif
    
    if (OSKSoundDriverDebugLevel > 1)
        NSLog(@"snddriver_stream_setup: page size = %d, sampleSize = %d", NSPageSize(), bytesPerSample);
    rc = snddriver_stream_setup([_locked_devicePort machPort], [_locked_ownerPort machPort],
                                streamType,
                                NSPageSize()/2, bytesPerSample,
                                0x40000, 0x80000,
                                &protocol, &streamPort);
    if (rc != KERN_SUCCESS) {
        NSZoneFree(NULL, _locked_recordQueue);
        _locked_recordQueue = NULL;
        [_locked_devicePort release];
        _locked_devicePort = nil;
        [_locked_ownerPort release];
        _locked_ownerPort = nil;
        
        [_doneLock unlock];
        mutex_unlock(_mutex);
        [pool release];
        [NSException raise: OSKSoundOutException
                    format: @"snddriver_stream_setup failed, rc = %d (%s)", rc, SNDSoundError(rc)];
    }
    _locked_streamPort = [[NSPort alloc] initWithMachPort: streamPort];

    rc = snddriver_set_sndout_bufcount([_locked_devicePort machPort], [_locked_ownerPort machPort], _numberOfBuffers);
    if (rc != KERN_SUCCESS)
        NSLog(@"snddriver_set_sndout_bufcount failed, rc = %d (%s)", rc, SNDSoundError(rc));

    if (OSKSoundDriverDebugLevel > 1)
        NSLog(@"OSKSoundOutSampleDestination: snddriver set up -- entering play state");

    _locked_playState = OSKPlayingPlayState;
    [NSThread detachNewThreadSelector: @selector(_play) toTarget: self withObject: nil];

    mutex_unlock(_mutex);

    // Do this outside the lock since we may well be in the main thread here
    // and this would then block.
    if ([_delegate respondsToSelector: @selector(playStateChanged)])
        [_delegate mainThreadPerformSelector: @selector(playStateChanged)];

    [pool release];
}


- (void) stop;
{
    kern_return_t rc;
    SNDError      sndError;

    mutex_lock(_mutex);

    if (_locked_playState == OSKStoppedPlayState) {
        mutex_unlock(_mutex);
        return;
    }

    rc = snddriver_stream_control([_locked_streamPort machPort], 0 /* tag */, SNDDRIVER_ABORT_STREAM);
    if (rc != KERN_SUCCESS)
        NSLog(@"-[%@ %@] -- snddriver_stream_control(SNDDRIVER_ABORT_STREAM) failed with code = %d",
              isa, NSStringFromSelector(_cmd), rc);

    while (_locked_pauseCount)
        // -resume will not reset the play state if it is 'stopped'
        [self resume];

    sndError = SNDRelease(SND_ACCESS_OUT, [_locked_devicePort machPort], [_locked_ownerPort machPort]);
    if (sndError != SND_ERR_NONE)
        NSLog(@"SNDRelease failed, sndError = %d, %s", sndError, SNDSoundError(sndError));

    [_locked_devicePort release];
    _locked_devicePort = nil;
    [_locked_ownerPort release];
    _locked_ownerPort = nil;
    [_locked_streamPort release];
    _locked_streamPort = nil;
    _locked_playState = OSKStoppedPlayState;
    if ([_delegate respondsToSelector: @selector(playStateChanged)])
        [_delegate mainThreadPerformSelector: @selector(playStateChanged)];
    [_locked_sampleStream release];
    _locked_sampleStream = nil;
    NSZoneFree(NULL, _locked_recordQueue);
    _locked_recordQueue = NULL;
    _locked_nextFinishingQueueEntry = 0;
    _locked_nextTag = 1;

    _nextTick = 0;
    _lastCompletedTick = 0;

    _insertSamplesAdjustmentCount = 0;
    _removeSamplesAdjustmentCount = 0;


    // Let those in -finish unblock
    [_doneLock unlock];

    mutex_unlock(_mutex);
}

/*" If the sound-out is not stopped, sets the volume.  Returns NO if the volume was not set. "*/
- (BOOL) setLeftVolume: (int) left  rightVolume: (int) right;
{
   BOOL didSet = NO;
    
    mutex_lock(_mutex);
    
    if (_locked_devicePort) {
        kern_return_t rc;

        fprintf(stderr, "Setting volume to %d/%d\n", left, right);
        rc = snddriver_set_volume([_locked_devicePort machPort], left, right);
        if (rc != KERN_SUCCESS)
            mach_error("snddriver_set_volume", rc);
        else
            didSet = YES;
    }

    mutex_unlock(_mutex);

    return didSet;
}

- (BOOL) getLeftVolume: (int *) left rightVolume: (int *) right;
{
    BOOL didGet = NO;

    mutex_lock(_mutex);

    if (_locked_devicePort) {
        kern_return_t rc;

        rc = snddriver_get_volume([_locked_devicePort machPort], left, right);
        if (rc != KERN_SUCCESS)
            mach_error("snddriver_get_volume", rc);
        else {
            didGet = YES;
            fprintf(stderr, "Got volume of %d/%d\n", *left, *right);
        }
    }

    mutex_unlock(_mutex);

    return didGet;
}

- (void) seekToSample: (OSKSampleIndex) sampleIndex;
{
#if 0
#warning We need another optional protocol that streams can implement to support seeking.
    mutex_lock(_mutex);
    if (sampleIndex < [_locked_sampleSource sampleCount]) {
        _locked_nextSampleIndex = sampleIndex;
    }
    mutex_unlock(_mutex);
#endif
}

- (OSKTick) nextTick;
{
    return _nextTick;
}

- (OSKTick) lastCompletedTick;
{
    return _lastCompletedTick;
}

- (void) pause;
{
#warning This needs to be disallowed in time-synchronized mode
    [_pauseLock lock];
    _locked_pauseCount++;
    if (_locked_playState == OSKPlayingPlayState) {
        kern_return_t rc;

        _locked_playState = OSKPausedPlayState;
        if ([_delegate respondsToSelector: @selector(playStateChanged)])
            [_delegate mainThreadPerformSelector: @selector(playStateChanged)];

        // Pause the sound driver.  This will cause the play thread to pause naturally
        // due to its outstanding buffer slot table filling up.
        // The tag is not useful for us -- it would be passed to our 'paused' or 'resumed' callback.
        rc = snddriver_stream_control([_locked_streamPort machPort], 0 /* tag */, SNDDRIVER_PAUSE_STREAM);
        if (rc != KERN_SUCCESS)
            NSLog(@"snddriver_stream_control failed with code = %d", rc);
    }
    [_pauseLock unlock];
}

- (void) resume;
{
#warning This needs to be disallowed in time-synchronized mode
    [_pauseLock lock];
    if (_locked_pauseCount) {
        _locked_pauseCount--;
        if (_locked_playState == OSKPausedPlayState) {
            kern_return_t rc;

            _locked_playState = OSKPlayingPlayState;
            if ([_delegate respondsToSelector: @selector(playStateChanged)])
                [_delegate mainThreadPerformSelector: @selector(playStateChanged)];

            // The tag is not useful for us -- it would be passed to our 'paused' or 'resumed' callback.
            rc = snddriver_stream_control([_locked_streamPort machPort], 0 /* tag */, SNDDRIVER_RESUME_STREAM);
            if (rc != KERN_SUCCESS)
                NSLog(@"snddriver_stream_control failed with code = %d", rc);
        }
    } else {
        // Raise an exception (outside of the lock of course) ?
    }
    [_pauseLock unlock];
}

- (unsigned int) pauseCount;
{
    return _locked_pauseCount;
}

- (OSKPlayState) playState;
{
    return _locked_playState;
}

- (void) finish;
{
    if (OSKSoundDriverDebugLevel > 0)
        NSLog(@"Waiting for sound to finish...");
    [_doneLock lock];
    [_doneLock unlock];
    if (OSKSoundDriverDebugLevel > 0)
        NSLog(@"Finished.");
}

@end


// This involves the latency in the sound driver, and the length of a buffer
#define COMPLETION_SLOP (0.5)


@implementation OSKSoundOutSampleDestination (PrivateAPI)

- (void) _locked_releaseRecord: (OSKBufferRecord *) bufferRecord;
{
    bufferRecord->tag = 0;
    bufferRecord->completionInterval     = 0.0;  // This will be valid at Jan 1, 2000, but the zero tag marks this as invalid
    bufferRecord->estimatedStartInterval = 0.0;  // This will be valid at Jan 1, 2000, but the zero tag marks this as invalid
    bufferRecord->duration = 0.0;
    [bufferRecord->sampleBuffer release];
    bufferRecord->sampleBuffer = nil;
    bufferRecord->isSilence  = NO;

    bufferRecord->hasStarted = NO;
    bufferRecord->actualStartInterval = 0.0;  // This will be valid at Jan 1, 2000, but the NO hasStarted marks this as invalid
}

- (BOOL) _locked_waitForAvailableQueuePosition;
{
    NSAutoreleasePool *pool;
    OSKBufferRecord   *bufferRecord;
    NSTimeInterval     now;
    BOOL               ret;

    pool = [[NSAutoreleasePool alloc] init];

    while (YES) {
        NSTimeInterval waitInterval;

        // See if we are supposed to continue.  The pause state will be taken care
        // of by the fact that we'll pause the sound driver itself.  This will cause
        // the outstanding buffer slot table to fill up and we'll pause naturally.
        if (_locked_playState == OSKStoppedPlayState) {
            if (OSKSoundDriverDebugLevel > 2)
                NSLog(@"WAIT    -- Done waiting, play state is OSKStoppedPlayState.");
            ret = NO;
            break;
        }

        bufferRecord = (OSKBufferRecord *)_locked_recordQueue + _locked_nextFinishingQueueEntry;
        if (!bufferRecord->tag) {
            if (OSKSoundDriverDebugLevel > 2)
                NSLog(@"WAIT    -- Done waiting, tag is zero in entry %d.", _locked_nextFinishingQueueEntry);
            // already available
            ret = YES;
            break;
        }

        // see if it has expired
        now = [NSDate timeIntervalSinceReferenceDate];
        waitInterval = bufferRecord->completionInterval - now + COMPLETION_SLOP;
        if (waitInterval < 0) {
            if (OSKSoundDriverDebugLevel > 0)
                NSLog(@"WAIT    -- tag = %d expired.  Should have completed at t = %f, but t = %f, waitInterval = %f, skipping...",
                      bufferRecord->tag,
                      bufferRecord->completionInterval - _baseTimeInterval,
                      now - _baseTimeInterval,
                      waitInterval);
            [self _locked_releaseRecord: bufferRecord];
            // ret = YES;
            // break;
            continue; // Try another record
        }

        // wait for it to complete or expire
        waitInterval = MAX(2.0, waitInterval);
        if (OSKSoundDriverDebugLevel > 2)
            NSLog(@"WAIT    -- Waiting %f seconds for buffer with tag %d to finish.",
                  waitInterval, bufferRecord->tag);
        if (condition_wait_timeout(_condition, _mutex, waitInterval * 1000)) {
            if (OSKSoundDriverDebugLevel > 2)
                NSLog(@"WAIT    -- Woken up");
        } else {
            if (OSKSoundDriverDebugLevel > 2)
                NSLog(@"WAIT    -- Timed out");
        }
        if (OSKSoundDriverDebugLevel > 2)
            NSLog(@"WAIT    -- Woke up %f seconds later.", [NSDate timeIntervalSinceReferenceDate] - now);

        mutex_unlock(_mutex);
        cthread_yield();
        mutex_lock(_mutex);
    }

    [pool release];
    return ret;
}

#define COARSE_SLOP_SECONDS (0.1)
#define FINE_SLOP_SECONDS (0.01)

- (void) _play;
{
    BOOL shouldStop = NO;
    
    if (OSKSoundDriverDebugLevel > 0)
        NSLog(@"PLAY: -_play entered");

    // This thread needs to have priority to keep the sound driver
    // in the kernel from running out of data.
    if (UseFixedPriorityMode) {
        if ([NSThread enabledFixedPriorityMode])
            [NSThread maximizePriority];
    }

    while (!shouldStop) {
        NSAutoreleasePool  *pool;
        OSKSampleBuffer    *sampleBuffer;
        BOOL                isSilence = NO;
        NSTimeInterval      estimatedStartInterval, desiredStartInterval, startTimeOffset;
        NSTimeInterval      minimumSynchronizationOffset, now, deadline;
        OFFastMutableData  *newSampleData = nil;
        OSKTick             minimumTick;
        unsigned int        bufferCount;
        OSKSampleStream    *sampleStream, *resampledStream;

        mutex_lock(_mutex);

        if (![self _locked_waitForAvailableQueuePosition]) {
            shouldStop = YES;
            mutex_unlock(_mutex);
            continue;
        }

        pool = [[NSAutoreleasePool alloc] init];

        // Get a new stream if necessary
        if (!_locked_sampleStream || [_locked_sampleStream isAtEnd]) {
            [self _locked_getNewSampleStream];

            if (!_locked_sampleStream || [_locked_sampleStream isAtEnd]) {
                shouldStop = YES;
                [pool release];
                mutex_unlock(_mutex);
                continue;
            }
        }

        // Store this in a local so that it doesn't go away immediately if we get stopped.
        sampleStream = _locked_sampleStream;
        resampledStream = nil;
        
        // Figure out how long we can wait.  First, get the estimated time at which
        // we will run out of sound.
        estimatedStartInterval = [self _locked_nextBufferStartTimeInterval: NO bufferCount: &bufferCount];

        // Get the next sample buffer.  This could block for a while, so unlock our
        // lock so play completion notices will get through.  We have to retain the 
        // sample stream during this since otherwise it might get deallocated
        // due to someone stopping us.
        [sampleStream retain];
        mutex_unlock(_mutex);

        if (!bufferCount) {
            // If there is nothing enqueued, we can wait forever.
            deadline = [[NSDate distantFuture] timeIntervalSinceReferenceDate];
        } else {
            // We can't delay until exactly the estimated start interval.  If we didn't get anything
            // then we'd have a real underrun.  In the case that the radio has nothing enqueued, our
            // silence buffers would slowly get out of sync (since we'd enqueue 0.5 seconds, try to
            // wait 0.5 seconds, but really end up waiting longer).  This is easier (and better overall)
            // than resetting the starting tick on underflow buffers based on the current time.
            now = [NSDate timeIntervalSinceReferenceDate];
            if (estimatedStartInterval - now > 0.05) {
                deadline = estimatedStartInterval - 0.05;
            } else
                deadline = estimatedStartInterval;
        }

        if (OSKSoundDriverDebugLevel > 2)
            NSLog(@"PLAY: Requesting sample buffer within %f seconds",
                  deadline - [NSDate timeIntervalSinceReferenceDate]);


        // Compute a somewhat sloppy minimum on the data we are looking for.  Don't wrap
        // negatively since OSKTick is unsigned (and that would produce a huge minimum).
        minimumTick = _nextTick;
        if (minimumTick > OSKSecondsToTicks(COARSE_SLOP_SECONDS))
            minimumTick -= OSKSecondsToTicks(COARSE_SLOP_SECONDS);

        sampleBuffer = [sampleStream nextSampleBufferByInterval: deadline minimumStartingTick: minimumTick];

        if (OSKSoundDriverDebugLevel > 2)
            NSLog(@"First sample buffer: rate=%d channels=%d bps=%d\n",
                  [sampleBuffer samplingRate], [sampleBuffer numberOfChannels], [sampleBuffer bytesPerSample]);
        
        // Interpose a LPF if necessary.
        if (sampleBuffer) {            
            if ([sampleBuffer samplingRate] != OSK44100SamplingRate ||
                [sampleBuffer bytesPerSample] != sizeof(OSKSample) ||
                [sampleBuffer numberOfChannels] != 2) {
                resampledStream = [[OSKResamplingStream alloc]
                                 initSampleRate: OSK44100SamplingRate
                                 bytesPerSample: sizeof(OSKSample)
                                samplesPerFrame: 2
                                         source: sampleStream
                                  initialBuffer: sampleBuffer];
            }

            if (resampledStream)
                sampleBuffer = [resampledStream nextSampleBufferByInterval: deadline minimumStartingTick: minimumTick];
        }

        if (OSKSoundDriverDebugLevel > 2)
            NSLog(@"First sample buffer* rate=%d channels=%d bps=%d  [ss=%x(%s) rs=%x sb=%x(rc=%d)]\n",
                  [sampleBuffer samplingRate], [sampleBuffer numberOfChannels],
                  [sampleBuffer bytesPerSample],
                  (int)sampleStream, ((id)sampleStream)->isa->name, (int)resampledStream, (int)sampleBuffer, [sampleBuffer retainCount]);
        
        // Relock
        mutex_lock(_mutex);

        if (resampledStream) {
            if (sampleStream == _locked_sampleStream) {
                [_locked_sampleStream release];
                _locked_sampleStream = resampledStream;
            } else {
                /* while we were waiting, someone gave us a new stream */
                [resampledStream release];
            }
        }
        
        [sampleStream release];

        if (!sampleBuffer) {
            // We didn't get anything!  Stick some zeros in the pipe to keep us happy.
            // We'll enqueue a fairly long buffer here (0.5 seconds).  That will both
            // reduce CPU usage during a stall and give the buffer provider time to recover.
            sampleBuffer = [[[OSKSampleBuffer alloc] initWithSampleData: underflowZeroData
                                                           samplingRate: OSK44100SamplingRate
                                                       numberOfChannels: 2
                                                         bytesPerSample: sizeof(OSKSample)
                                                           startingTick: _nextTick] autorelease];
            isSilence = YES;

            if (OSKSoundDriverDebugLevel > 0)
                NSLog(@"PLAY: Got nothing at all -- inserting zero sampleBuffer = %@, startingTick = %d",
                      sampleBuffer, (unsigned int)[sampleBuffer startingTick]);
        }

        // We might have waited quite a while if there was nothing previously enqueued.
        // Recompute the estimated starting time.
        if (!bufferCount)
            estimatedStartInterval = [self _locked_nextBufferStartTimeInterval: NO bufferCount: &bufferCount];

        if ([sampleBuffer startingTick] == OSK_INVALID_TICK)
            desiredStartInterval = estimatedStartInterval;
        else

            desiredStartInterval = OSKTicksToSeconds([sampleBuffer startingTick]) + _baseTimeInterval;
        startTimeOffset = desiredStartInterval - estimatedStartInterval;
        if (OSKSoundDriverDebugLevel > 2)
            NSLog(@"PLAY: desiredStartInterval = %g, estimatedStartInterval = %g, startTimeOffset = %f",
                  desiredStartInterval - _baseTimeInterval, estimatedStartInterval - _baseTimeInterval, startTimeOffset);

        if (!isSilence) {
            if (fabs(startTimeOffset) > COARSE_SLOP_SECONDS) {
                // Need to resynchronize.  For the time being, we'll do this the ugly sounding way
                // of either inserting blank space or dropping packets.

                if (startTimeOffset > 0) {
                    // We need to insert some blank space
                    OFFastMutableData  *newSamples, *oldSamples;
                    OSKTick             silentTicks, startingTick;
                    OSKSampleIndex      silentSamples;
                    unsigned            silentByteCount;
                    
                    silentTicks = OSKSecondsToTicks(startTimeOffset);
                    silentSamples = OSKTicksToSamples(silentTicks, OSK44100SamplingRate);

                    if (OSKSoundDriverDebugLevel > 0)
                        NSLog(@"PLAY: ####### Prepending %f seconds of silence (%d samples). #######",
                              startTimeOffset, (unsigned int)silentSamples);

                    if (startTimeOffset > 10.0) {
                        // TJW -- My machine crashes if I try to enqueue too much silence
                        NSLog(@"PLAY: Too much silence.");
                        shouldStop = YES;
                        mutex_unlock(_mutex);
                        continue;
                    }

                    // Get a data big enough for the silence plus the stuff we really want
                    // to enqueue.
                    silentByteCount = silentSamples * 2 * sizeof(OSKSample);
                    oldSamples = [sampleBuffer sampleData];
                    newSamples = [OFFastMutableData newFastMutableDataWithLength:
                        silentByteCount + [oldSamples length]];

                    // Zero out the beginning
                    memset([newSamples mutableBytes], 0, silentByteCount);

                    // Copy over the real samples
                    memcpy([newSamples mutableBytes] + silentByteCount, [oldSamples bytes], [oldSamples length]);
                           
                    // Generate a new sample buffer combined from the two.  Have to keep the same starting
                    // tick so that we don't jump back to a desired starting tick of zero on the first
                    // buffer (which would never work unless the _baseTimeInterval was exactly now
                    // or something like that).  But, since the silence should start before this
                    // buffer wanted to start, we should subtract the silentTicks.  This will give
                    // the first non-silent sample the same starting tick as it wanted.
                    startingTick = [sampleBuffer startingTick] - silentTicks;
                    sampleBuffer = [[[OSKSampleBuffer alloc] initWithSampleData: newSamples
                                                                   samplingRate: OSK44100SamplingRate
                                                               numberOfChannels: 2
                                                                 bytesPerSample: sizeof(OSKSample)
                                                                   startingTick: startingTick] autorelease];
                    [newSamples release];

                    if (OSKSoundDriverDebugLevel > 0)
                        NSLog(@"PLAY: ####### New duration = %f, sampleData length = %d #######",
                              [sampleBuffer durationSeconds], [[sampleBuffer sampleData] length]);

                } else {
                    OSKTick newNextTick;
                    
                    // For now, we'll simply drop this packet.  This will likely mean that
                    // on the next loop we'll need to insert some blank space (if the start
                    // of the packet should have already been enqueued but the end not).
                    // Later, we'll could chop off the beginning of the buffer.  That will
                    // still be ugly sounding so it isn't clear that it is worth while.
                    if (OSKSoundDriverDebugLevel > 0)
                        NSLog(@"PLAY: Dropping buffer, desiredStartInterval = %g, estimatedStartInterval = %g, startTimeOffset = %f",
                              desiredStartInterval - _baseTimeInterval, estimatedStartInterval - _baseTimeInterval, startTimeOffset);
                    sampleBuffer = nil;

                    if (OSKSoundDriverExitRatherThanResettingNextTick)
                        exit(1);
                    // Also, this means that the stream has old stale data.  Update the _nextTick ivar
                    // so that it will purge stale data.
                    newNextTick = OSKSecondsToTicks((estimatedStartInterval + 1.0) - _baseTimeInterval);
                    if (OSKSoundDriverDebugLevel > 0)
                        NSLog(@"PLAY: Resetting nextTick from %d to %d.", (unsigned int)_nextTick, (unsigned int)newNextTick);
                    _nextTick = newNextTick;
                }

            // Only do fine adjustments every 2 times through the buffer, to give our last adjustment a chance to get played and affect all the buffer offsets.
            } else if (_locked_nextTag % ( _numberOfBuffers * 2) == 0) {
                // Look at the average difference between the actual and real start intervals.
                // Due to scheduling delays, this isn't going to be an exact, but it should
                // make a reasonable average.  The one problem with this is that if we underflow
                // we can get big ugly numbers in the delay.  Thus, we'll only use the average
                // to determine which direction to adjust and we'll only adjust a small amount
                // at a time (so that the user can't easily hear the difference).

                // The number of samples we remove or insert should be small compared to
                // the sampling rate to avoid clicks.  We'll start with 4 samples per
                // adjustment (1/10000th of a second when at 44.1kHz).

                // This is the average of the (desired - actual) start intervals.
                minimumSynchronizationOffset = [self _locked_minimumSynchronizationOffset];
                if (minimumSynchronizationOffset > FINE_SLOP_SECONDS) {
                    OSKSample    *samples;
                    unsigned int sampleCount;

                    _insertSamplesAdjustmentCount++;
                    
                    // The desired time is, on average, greater than the actual start time.
                    // This means that we are starting too soon and we should insert some samples.
                    newSampleData = [sampleBuffer sampleData];
                    OBASSERT(newSampleData);
                    sampleCount = [newSampleData length] / (2 * sizeof(OSKSample));

                    // we need to clean up terminology -- this is 4 stereo pairs
                    [newSampleData setLength: (sampleCount + 8) * (2 * sizeof(OSKSample))];
                    samples = [newSampleData mutableBytes];
                    OBASSERT(samples);

                    // For now, we'll copy the last sample four times.  We could attempt
                    // smooth it out more, but it isn't clear that this would be enough
                    // better to be worth the effort.
                    samples[sampleCount+0] = samples[sampleCount-2];  // left0
                    samples[sampleCount+1] = samples[sampleCount-1];  // right0
                    samples[sampleCount+2] = samples[sampleCount-2];  // left1
                    samples[sampleCount+3] = samples[sampleCount-2];  // right1
                    samples[sampleCount+4] = samples[sampleCount-2];  // left2
                    samples[sampleCount+5] = samples[sampleCount-2];  // right2
                    samples[sampleCount+6] = samples[sampleCount-2];  // left3
                    samples[sampleCount+7] = samples[sampleCount-2];  // right3

                    if (OSKSoundDriverDebugLevel > 1)
                        NSLog(@"PLAY: minimumSynchronizationOffset = %f, adding samples", minimumSynchronizationOffset);
                } else if (minimumSynchronizationOffset < -FINE_SLOP_SECONDS) {
                    unsigned int sampleCount;

                    _removeSamplesAdjustmentCount++;
                    
                    // The desired time is, on average, greater than the actual start time.
                    // This means that we are starting too late and we should drop some samples.

                    newSampleData = [sampleBuffer sampleData];
                    sampleCount = [newSampleData length] / (2 * sizeof(OSKSample));

                    // we need to clean up terminology -- this is 4 stereo pairs
                    if (sampleCount > 8)
                        // abort if we aren't long enough
                        [newSampleData setLength: (sampleCount - 8) * (2 * sizeof(OSKSample))];
                    if (OSKSoundDriverDebugLevel > 1)
                        NSLog(@"PLAY: minimumSynchronizationOffset = %f, dropping samples", minimumSynchronizationOffset);
                }

                if (newSampleData) {
                    // The desired start time stays the same (the whole point is to leave that the same
                    // and change the *actual* start time), so we don't need to adjust that at all
                    // just the length changes.
                    sampleBuffer = [[OSKSampleBuffer alloc] initWithSampleData: newSampleData
                                                                  samplingRate: [sampleBuffer samplingRate]
                                                              numberOfChannels: [sampleBuffer numberOfChannels]
                                                                bytesPerSample: [sampleBuffer bytesPerSample]
                                                                  startingTick: [sampleBuffer startingTick]];
                    [sampleBuffer autorelease];
                    if (OSKSoundDriverDebugLevel > 1) {
                        NSLog(@"PLAY: new sampleBuffer duration = %f", [sampleBuffer durationSeconds]);
                    }
                }
            }
        }


#if 0
        NSLog(@"sb=%x\n", (int)sampleBuffer);
        [sampleBuffer retain];
        [self retain];
        [pool release]; pool = (id)~0;
        pool = [[NSAutoreleasePool alloc] init];
        [sampleBuffer autorelease];
        [self autorelease];
#endif

        if (sampleBuffer) {
            [self _locked_enqueueSampleBuffer: sampleBuffer
                       estimatedStartInterval: estimatedStartInterval
                                    isSilence: isSilence];

            if ([self _locked_containsOnlySilence: NO]) {
                // We are full of silence.  There really isn't much point in playing any longer.
                // Also, our silence streaming stuff still gets out of sync (since there is no
                // guarantee we'll get the next silence buffer enqueued before the previous one
                // runs out, and we don't adjust the starting tick based on the current time).
                shouldStop = YES;
            }
        }

        [pool release]; pool = (id)~0;
        mutex_unlock(_mutex);

        // Don't monopolize the CPU enqueuing sound since the sound might be
        // generated dynamically, or we might need to service the network really
        // soon.
        cthread_yield();
    }

    if (OSKSoundDriverDebugLevel > 0)
        NSLog(@"PLAY: -_play loop exiting");

    // It is difficult to just write something that waits for our table of outstanding
    // buffers to finish.  Instead, we'll enqueue buffers full of zeros until only
    // silence is enqueued.  Due to our locking strategy, this will ensure that all of the
    // interesting buffers are done.
    shouldStop = NO;
    while (!shouldStop) {
        mutex_lock(_mutex);

        if ([self _locked_containsOnlySilence: NO]) {
            shouldStop = YES;
        } else if (![self _locked_waitForAvailableQueuePosition]) {
            // oops.  someone really wants us to cut it out
            shouldStop = YES;
        } else {
            [self _locked_enqueueSampleBuffer: zeroSampleBuffer estimatedStartInterval: 0.0 isSilence: YES];
        }
        
        mutex_unlock(_mutex);
    }

    // Now we can kill the stream since it only contains silence
    [self stop];
}

/*"
Schedules a buffer to be played as soon as possible (ie, right after the previous buffer).  No actual synchronization changes are to be made in this code although this method collects some data necessary for the higher level constructs.
"*/
- (void) _locked_enqueueSampleBuffer: (OSKSampleBuffer *) sampleBuffer
              estimatedStartInterval: (NSTimeInterval) estimatedStartInterval
                           isSilence: (BOOL) isSilence;
{
    NSData            *sampleData;
    unsigned int       sampleBufferFrameCount; // not an OSKSampleIndex since we're passing it to the snddriver.
    unsigned int       sampleBufferSampleCount; // not an OSKSampleIndex since we're passing it to the snddriver.
    kern_return_t      rc;
    OSKBufferRecord   *bufferRecord;
    NSTimeInterval     now, endInterval;
    OSKTick            desiredStartingTick, endingTick;

    if (!_locked_recordQueue)
        // We've apparently been stopped
        return;
    
    sampleData = [sampleBuffer sampleData];
    sampleBufferFrameCount = [sampleBuffer numberOfSamples];
    now = [NSDate timeIntervalSinceReferenceDate];

    endInterval   = estimatedStartInterval + [sampleBuffer durationSeconds];

    
    bufferRecord = (OSKBufferRecord *)_locked_recordQueue + _locked_nextFinishingQueueEntry;
    bufferRecord->tag = _locked_nextTag++;
    bufferRecord->completionInterval = endInterval;
    bufferRecord->sampleBuffer = [sampleBuffer retain];
    bufferRecord->estimatedStartInterval = estimatedStartInterval;
    bufferRecord->duration = [sampleBuffer durationSeconds];
    desiredStartingTick = [sampleBuffer startingTick];
    bufferRecord->desiredStartInterval = (desiredStartingTick == OSK_INVALID_TICK)?bufferRecord->estimatedStartInterval:(OSKTicksToSeconds(desiredStartingTick)+_baseTimeInterval);
    bufferRecord->isSilence = isSilence;
    
    sampleBufferSampleCount = sampleBufferFrameCount * [sampleBuffer numberOfChannels]; // account for stereo

    if (OSKSoundDriverDebugLevel > 2)
        NSLog(@"PLAY    -- entry = %d, tag = %d, desired start = %f, completing at t = %f, duration = %f",
              _locked_nextFinishingQueueEntry, bufferRecord->tag,
              bufferRecord->desiredStartInterval - _baseTimeInterval,
              bufferRecord->completionInterval - _baseTimeInterval,
              bufferRecord->duration);

    // Make sure we can get messages back from the kernel while sending message to it
    mutex_unlock(_mutex);
    rc = snddriver_stream_start_writing([_locked_streamPort machPort], (void *)[sampleData bytes],
                                        sampleBufferSampleCount, bufferRecord->tag,
                                        preempt, deallocateWhenDone,
                                        msgStarted, msgCompleted,
                                        msgAborted, msgPaused,
                                        msgResumed, msgUnderrun,
                                        [replyPort machPort]);
    mutex_lock(_mutex);
    
    if (rc != KERN_SUCCESS)
        NSLog(@"PLAY    -- snddriver_stream_start_writing failed with code = %d", rc);

    // Set our guess as to the next tick.  Don't set it unless it is greater than our last
    // guess.  This is necessary since the end of the _play method enqueues buffers with a
    // starting tick of zero.  When recovering from a queue underrun, we'd end up specifying
    // a minimum of something really low otherwise (negative in fact, confusing the receiver
    // into thinking it was a huge number).
    endingTick = [bufferRecord->sampleBuffer endingTick];
    if (_nextTick < endingTick)
        _nextTick = endingTick;

    _locked_nextFinishingQueueEntry = (_locked_nextFinishingQueueEntry + 1) % _numberOfBuffers;
}

/*"
Returns YES if there are no outstanding non-silent buffers.  If andIsFull is set, this will only return YES if all of the entries are full.
 "*/
- (BOOL) _locked_containsOnlySilence: (BOOL) andIsFull;
{
    unsigned int   bufferIndex, bufferCount;

    if (!_locked_recordQueue)
        // We've been stopped (and all of the entries that exist -- none -- are full).
        return YES;
    
    bufferCount = _numberOfBuffers;
    for (bufferIndex = 0; bufferIndex < bufferCount; bufferIndex++) {
        OSKBufferRecord *record;

        record = &((OSKBufferRecord *)_locked_recordQueue)[bufferIndex];

        if (andIsFull && !record->tag)
            return NO;
        
        if (record->tag && !record->isSilence)
            return NO;
    }

    return YES;
}

/*"
Returns the average time interval that the start of each recent packet differed from its estimated start time.  Should only be called privately when we are locked.
"*/
- (NSTimeInterval) _locked_minimumSynchronizationOffset;
{
    unsigned int   bufferIndex;
    NSTimeInterval minimumSynchronizationOffset;

    if (!_locked_recordQueue)
        // We're not playing, or something like that
        return 0.0;

    minimumSynchronizationOffset = 32768.0;

    for (bufferIndex = 0; bufferIndex < _numberOfBuffers; bufferIndex++)
        minimumSynchronizationOffset = MIN(minimumSynchronizationOffset, ((OSKBufferRecord *)_locked_recordQueue)[bufferIndex].lastStartOffset);

    return minimumSynchronizationOffset;
}

+ (void) _readReplyMessages;
{
    /* pass our slot manager as one of the arguments to the snd handlers */
    snddriver_handlers_t handlers = {
        NULL, // user data
        0,
        play_started,
        play_completed,
        play_aborted,
        NULL, // paused
        NULL, // resumed
        play_overflow, // overflow
        NULL, // recorded_data
        NULL, // condition_true
        NULL, // dep_message
        NULL  // dsp_error
    };
    kern_return_t      rc;
    msg_header_t      *replyMessage;
    port_t             localReplyPort;
    NSAutoreleasePool *pool;
    
    // Read the reply port once and cache it after releasing the lock.  The only
    // case we care about is when our port gets deallocated (indicating that we
    // should quit).  msg_receive will let us know when this happens.
    localReplyPort = [replyPort machPort];
    replyMessage = NSZoneMalloc(NULL, MSG_SIZE_MAX);

    // This thread needs to have priority to keep the sound driver
    // in the kernel from blocking on us and underrunning.
    pool = [[NSAutoreleasePool alloc] init];
    if ([[OFUserDefaults sharedUserDefaults] boolForKey: @"UseFixedPriorityMode"]) {
        if ([NSThread enabledFixedPriorityMode])
            [NSThread maximizePriority];
    }
    [pool release];
    
    while (YES) {
        replyMessage->msg_size = MSG_SIZE_MAX;
        replyMessage->msg_local_port = localReplyPort;
        rc = msg_receive(replyMessage, MSG_OPTION_NONE, 0);
        switch (rc) {
            case RCV_SUCCESS:
                rc = snddriver_reply_handler(replyMessage, &handlers);
                if (rc != KERN_SUCCESS) {
                    NSLog(@"snddriver_reply_handler failed with code = %d", rc);
                }
                break;
            case RCV_INVALID_PORT:
                // If our port gets deallocated, that's our signal to finish up.
                // Actually, this shouldn't happen anymore now that we share
                // a reply thread for the whole class forever.
                return;
            default:
                mach_error("_readReplyMessages - msg_receive", rc);
                break;
        }
    }

    NSZoneFree(NULL, replyMessage);
}

- (void) _locked_getNewSampleStream;
{
    OSKSampleStream *newSampleStream;

    if (![_delegate respondsToSelector:@selector(nextSampleStream)])
        return;

    newSampleStream = [_delegate nextSampleStream];
    if (newSampleStream == _locked_sampleStream)
        return;

    [self _locked_setNewSampleStream:newSampleStream];

    if (OSKSoundDriverDebugLevel > 1)
        NSLog(@"_locked_getNewSampleStream: _locked_sampleStream = %@", _locked_sampleStream);
}

- (void) _locked_setNewSampleStream:(OSKSampleStream *)newStream
{
    [_locked_sampleStream release];
    _locked_sampleStream = nil;
    if (newStream) {
        _locked_sampleStream = [newStream retain];
    }
}

/*"
Figures out an estimate of when a sound buffer we enqueue would actually start.  Thus, if there are no buffers enqueued, will return the current time.

If there are any buffers enqueued, we will use the actual start time of the first buffer to determine how much of it has completed.  Adding the durations of the remaining buffers will give us a good estimate of when a new buffer would actually start playing.

This estimate, when used to determine how long can be spent in obtaining new samples, is not always terribly conservative.  It does not count overhead of the work of building and enqueuing the buffer.  Higher level interfaces must build in this safety.

If fromNow is YES, this returns the ofset from the current time that a new buffer would start.  That is, it returns the duration of the enqueued sound.  Otherwise, it returns the absolute time interval at which a new buffer should start.
"*/
- (NSTimeInterval) _locked_nextBufferStartTimeInterval: (BOOL) fromNow
                                           bufferCount: (unsigned int *) bufferCount;
{
    OSKBufferRecord *bufferRecord;
    unsigned int     bufferIndex;
    NSTimeInterval   now;
    NSTimeInterval   startInterval;
    unsigned int     queuedBufferCount = 0;
    
    // A buffer cannot start any sooner than immediately.  If we could figure out
    // how to do that, we'd all be fabulously weathly.
    now = [NSDate timeIntervalSinceReferenceDate];
    if (fromNow)
        startInterval = 0.0;
    else
        startInterval = now;

    if (!_locked_recordQueue) {
        *bufferCount = queuedBufferCount;
        return startInterval;
    }
    
    // Loop over all of the queue entries.  If an entry is pending but not currently
    // playing, add in the full duration of the entry.  If the entry has started
    // playing, prorate the duration based on the difference between its start
    // time and the current time.
    for (bufferIndex = 0; bufferIndex < _numberOfBuffers; bufferIndex++) {
        bufferRecord = (OSKBufferRecord *)_locked_recordQueue + bufferIndex;
        if (!bufferRecord->tag)
            // Nothing in this slot
            continue;

        queuedBufferCount++;
        
        if (!bufferRecord->hasStarted) {
            // Add in the full time
            startInterval += bufferRecord->duration;
            if (OSKSoundDriverDebugLevel > 2)
                NSLog(@"CALC START: entry = %d, tag = %d, duration = %f", bufferIndex, bufferRecord->tag, bufferRecord->duration);
        } else {
            // Prorate the duration
            NSTimeInterval durationPlayed, fineAdjustment;

            durationPlayed = now - bufferRecord->actualStartInterval;
            fineAdjustment = bufferRecord->duration - durationPlayed;

            if (OSKSoundDriverDebugLevel > 2)
                NSLog(@"CALC START: entry = %d, tag = %d, fineAdjustment = %f, now = %f, actualStartInterval = %f/%f, durationPlayed = %f, totalDuration = %f",
                      bufferIndex, bufferRecord->tag, fineAdjustment,
                      now - _baseTimeInterval, bufferRecord->actualStartInterval, bufferRecord->actualStartInterval - _baseTimeInterval,
                      durationPlayed, bufferRecord->duration);
            startInterval += fineAdjustment;
        }
    }

    if (OSKSoundDriverDebugLevel > 2)
        NSLog(@"CALC START: %d buffers queued, startInterval = %f  (%f - %f)",
              queuedBufferCount, startInterval - _baseTimeInterval, startInterval, _baseTimeInterval);

    *bufferCount = queuedBufferCount;
    return startInterval;
}

/*"
Mark down the time that the buffer actually started playing.  This will allow us to make a much better estimate of when it will end inside of -_locked_nextBufferStartTimeInterval:bufferCount:.
"*/
- (void) _playStarted: (int) tag;
{
    OSKBufferRecord *bufferRecord;
    unsigned int     queueIndex;
    NSTimeInterval   now, offset;

    // Get this outside of the lock for better accuracy
    now = [NSDate timeIntervalSinceReferenceDate];

    mutex_lock(_mutex);

    if (!_locked_recordQueue) {
        // We might have been stopped while waiting for the lock
        goto found;
    }
        
    // Find the buffer record.  There should only be one entry with this tag.
    for (queueIndex = 0; queueIndex < _numberOfBuffers; queueIndex++) {
        bufferRecord = (OSKBufferRecord *)_locked_recordQueue + queueIndex;
        if (bufferRecord->tag == tag) {
            bufferRecord->actualStartInterval = now;
            bufferRecord->hasStarted = YES;

            offset = bufferRecord->desiredStartInterval - now;
            if (OSKSoundDriverDebugLevel > 1)
                NSLog(@"DID START: Play started on tag %d, offset = %f (desired = %f, now = %f, base = %f)",
                      tag, offset, bufferRecord->desiredStartInterval, now, _baseTimeInterval);

            // We compute a recent average for the last _numberOfBuffers buffers.
            if (tag == 1) {
                unsigned int bufferIndex;

                // Initialize all buffers to this level so we get a reasonable starting average
                for (bufferIndex = 0; bufferIndex < _numberOfBuffers; bufferIndex++)
                    ((OSKBufferRecord *)_locked_recordQueue)[bufferIndex].lastStartOffset = offset;
            } else {
                // Update just the last latency for this buffer
                bufferRecord->lastStartOffset = offset;
            }

            goto found;
        }
    }

    if (OSKSoundDriverDebugLevel > 0)
        NSLog(@"START -- No buffer with tag %d found", tag);
    
found:
    mutex_unlock(_mutex);
}

- (void) _playCompleted: (int) tag;
{
    OSKBufferRecord *bufferRecord;
    unsigned int     bufferIndex;
    NSTimeInterval   now;

    mutex_lock(_mutex);

    if (!_locked_recordQueue) {
        // We must have stopped
        condition_signal_timeout(_condition);
        mutex_unlock(_mutex);
        return;
    }
    
    now = [NSDate timeIntervalSinceReferenceDate];

    /* Dequeue this buffer and any that should have come before it */
    for (bufferIndex = 0; bufferIndex < _numberOfBuffers; bufferIndex++) {
        bufferRecord = (OSKBufferRecord *)_locked_recordQueue + bufferIndex;

        if (bufferRecord->tag && bufferRecord->tag <= tag) {
            if (OSKSoundDriverDebugLevel > 2)
                NSLog(@"DID END: Dequeueing buffer with tag %d (dt = %f)",
                      bufferRecord->tag, bufferRecord->completionInterval - now);
            [self _locked_releaseRecord: bufferRecord];
        }
    }
    
    condition_signal_timeout(_condition);
    mutex_unlock(_mutex);
}

@end



static void play_aborted(void *userData, int tag)
{
    if (OSKSoundDriverDebugLevel > 0)
        NSLog(@"Play aborted on tag %d", tag);
}

static void play_started(void *userData, int tag)
{
    [sharedSoundOutSampleDestination _playStarted: tag];
}

static void play_completed(void *userData, int tag)
{
    [sharedSoundOutSampleDestination _playCompleted: tag];
}

static void play_overflow(void *userData, int tag)
{
    fprintf(stderr, "** Over/Underflow\n");
}

